import enum
import logging

import numpy as np
import typing as ty

import armarx
from armarx_memory.aron.aron_ice_types import AronDataIceTypes


def pythonic_to_aron_ice(
    value: ty.Any,
) -> "armarx.aron.data.dto.GenericData":
    """
    Deeply converts objects/values of pythonic types to their Aron Ice counterparts.

    :param value: A pythonic object or value.
    :param options: Conversion options.
    :return: An Aron data Ice object.
    """
    from .ndarray.common import ndarray_to_aron

    if value is None:
        return None
    if isinstance(value, str):
        return AronDataIceTypes.string(value)
    elif isinstance(value, bool):
        return AronDataIceTypes.bool(value)
    elif isinstance(value, np.int64):
        return AronDataIceTypes.long(int(value))
    elif isinstance(value, int) or isinstance(value, np.int32):
        assert(isinstance(value, int)), value
        assert 'invalid value - expected int' not in str(AronIceTypes.int(int(value))), \
            f'Casting {value} to int failed. Did you intend to use np.int64 instead, ' \
            'but assigned a plain int value, or converted the value somewhere in the meantime?'
        return AronIceTypes.int(int(value))
    elif isinstance(value, np.float64):
        return AronDataIceTypes.double(float(value))
    elif isinstance(value, float) or isinstance(value, np.float32):
        return AronDataIceTypes.float(float(value))
    elif isinstance(value, list):
        return AronDataIceTypes.list(list(map(pythonic_to_aron_ice, value)))
    elif isinstance(value, enum.IntEnum):
        assert(isinstance(value.value, int)), f'Value of IntEnum is not an int: {value.value}'
        return pythonic_to_aron_ice(value.value)  # int
    elif isinstance(value, enum.Enum):
        if not isinstance(value.value, int):
            raise NotImplementedError(
                f'enum.Enum is only supported as long as it uses int values, but got value {value.value}. '
                'Consider directly using enum.IntEnum.'
            )
        return pythonic_to_aron_ice(value.value)

    elif isinstance(value, dict):
        a = AronDataIceTypes.dict({k: pythonic_to_aron_ice(v) for k, v in value.items()})
        return a

    elif isinstance(value, AronDataIceTypes.Dict):
        return value

    elif isinstance(value, np.ndarray):
        return ndarray_to_aron(value)

    try:
        return value.to_aron()
    except TypeError:
        pass

    raise TypeError(f"Could not convert object of type '{type(value)}' to aron.")


def pythonic_from_aron_ice(
    data: "armarx.aron.data.dto.GenericData",
    logger: ty.Optional[logging.Logger] = None,
) -> ty.Any:
    """
    Deeply converts an Aron data Ice object to its pythonic representation.

    :param data: The Aron data Ice object.
    :param logger: Logger for additional logging.
    :return: The pythonic representation.
    """
    from .ndarray.common import ndarray_from_aron

    def handle_dict(elements):
        return {k: pythonic_from_aron_ice(v) for k, v in elements.items()}

    def handle_list(elements):
        return list(map(pythonic_from_aron_ice, elements))

    if data is None:
        return None
    if isinstance(data, list):
        return handle_list(data)
    elif isinstance(data, dict):
        return handle_dict(data)

    elif isinstance(data, (float, int, str)):
        return data

    if isinstance(data, AronDataIceTypes.NDArray):
        return ndarray_from_aron(data)

    if isinstance(data, AronDataIceTypes.Long):
        return np.int64(data.value)

    if isinstance(data, np.int32):
        return data
    if isinstance(data, np.int64):
        return data

    try:
        return data.value
    except AttributeError:
        pass

    try:
        elements = data.elements
    except AttributeError:
        pass
    else:
        if isinstance(elements, list):
            return handle_list(elements)
        elif isinstance(elements, dict):
            return handle_dict(elements)
        else:
            raise TypeError(
                f"Could not handle aron container object of type '{type(data)}'. \n"
                f"elements: {elements}"
            )

    raise TypeError(
        f"Could not handle aron object of type '{type(data)}'.\n" f"dir(a): {dir(data)}"
    )
