from dataclasses import dataclass, field

import numpy as np

from armarx_memory.aron.data.variant import AronDict, AronFloat
from armarx_memory.aron.data.dto.aron_dto import AronDTOBase
from armarx_memory.aron.data.dto.common.instantiated_type import InstantiatedContainerType, InstantiatedLeafType


@dataclass
class Configuration(AronDTOBase):
    config: dict[str, np.float32] = field(default_factory=dict)

    @classmethod
    def fromAron(cls, d: AronDict):
        config = {}

        joints = d.at("joints")
        assert isinstance(joints, AronDict)

        positions = joints.at("position")
        assert isinstance(positions, AronDict)

        for name, jj in positions.elements.items():
            assert isinstance(jj, AronFloat)
            config[name] = jj.value
        
        return cls(config)
    
    def toAron(self) -> AronDict:
        root: AronDict = AronDict(path=[])
        joints: AronDict = AronDict(path=["joints"])
        position: AronDict = AronDict(path=["joints", "position"])

        for k, v in self.config.items():
            p: AronFloat = AronFloat(path=[], value=v)
            position.add(k, p)

        joints.add("position", position)
        root.add("joints", joints)
        return root

    def toNDArray(self) -> tuple[np.ndarray, InstantiatedContainerType]:
        array: np.ndarray = np.empty((0,1), dtype=np.float32)
        
        position_structure = InstantiatedContainerType(type=AronDict.IDENTIFICATOR, cols=len(self.config), elements={})
        for k, c in self.config.items():
            array = np.append(array, np.array([[c]], dtype=np.float32))
            position_structure.elements.update({
                k: InstantiatedLeafType(type=AronFloat.IDENTIFICATOR, cols=1)
            })
        
        joints_structure = InstantiatedContainerType(type=AronDict.IDENTIFICATOR, cols=position_structure.cols, elements={"position": position_structure})
        structure = InstantiatedContainerType(type=AronDict.IDENTIFICATOR, cols=joints_structure.cols, elements={"joints": joints_structure})

        return array, structure
     
    @classmethod
    def fromNDArray(cls, array: np.ndarray, type_structure: InstantiatedContainerType):
        assert len(array.shape) == 2
        assert array.shape[-1] == 1

        assert "joints" in type_structure.elements.keys()
        joints_structure = type_structure.elements["joints"]
        assert isinstance(joints_structure, InstantiatedContainerType)

        assert "position" in joints_structure.elements.keys()
        position_structure = joints_structure.elements["position"]
        assert isinstance(position_structure, InstantiatedContainerType)

        r: Configuration = Configuration()

        i = 0
        for k, leaf_structure  in position_structure.elements.items():
            assert isinstance(leaf_structure, InstantiatedLeafType)

            assert leaf_structure.type == AronFloat.IDENTIFICATOR
            assert leaf_structure.cols == 1
            r.config[k] = array[i,0]
            i+= 1

        return r