import struct
from dataclasses import dataclass, field
from typing import Any

import numpy as np


# prototype
@dataclass
class AronInt:
    pass
@dataclass
class AronLong:
    pass
@dataclass
class AronFloat:
    pass
@dataclass
class AronDouble:
    pass
@dataclass
class AronString:
    pass
@dataclass
class AronBool:
    pass
@dataclass
class AronNDArray:
    pass
@dataclass
class AronList:
    pass
@dataclass
class AronDict:
    pass


# typing alias
AronData = AronInt | AronLong | AronFloat | AronDouble | AronBool | AronString | AronNDArray | AronList | AronDict


# ARON primitive types
@dataclass
class AronInt(AronInt):
    IDENTIFICATOR = "int"

    path: list[str] = field(default_factory=list)
    value: np.int32 = field(default_factory=lambda: 0)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_VALUE": int(self.value),
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_INT"
        }
        return j

    def to_primitive(self):
        return int(self.value)

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_INT", "received wrong type for aron int"
        path: list[str] = j["_ARON_PATH"]
        value: np.int32 = np.int32(int(j["_ARON_VALUE"]))
        return cls(path=path, value=value)


@dataclass
class AronLong(AronLong):
    IDENTIFICATOR = "long"

    value: np.int64 = field(default_factory=lambda: 0)
    path: list[str] = field(default_factory=list)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_VALUE": int(self.value),
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_LONG"
        }
        return j

    def to_primitive(self):
        return int(self.value)

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_LONG", "received wrong type for aron long"
        path: list[str] = j["_ARON_PATH"]
        value: np.int64 = np.int64(int(j["_ARON_VALUE"]))
        return cls(path=path, value=value)


@dataclass
class AronFloat(AronFloat):
    IDENTIFICATOR = "float"

    value: np.float32 = field(default_factory=lambda: 0.0)
    path: list[str] = field(default_factory=list)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_VALUE": float(self.value),
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_FLOAT"
        }
        return j

    def to_primitive(self):
        return float(self.value)

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_FLOAT", "received wrong type for aron float"
        path: list[str] = j["_ARON_PATH"]
        value: np.float32 = np.float32(float(j["_ARON_VALUE"]))
        return cls(path=path, value=value)


@dataclass
class AronDouble(AronDouble):
    IDENTIFICATOR = "double"

    value: np.float64 = field(default_factory=lambda: 0.0)
    path: list[str] = field(default_factory=list)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_VALUE": float(self.value),
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_DOUBLE"
        }
        return j

    def to_primitive(self):
        return float(self.value)

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_DOUBLE", "received wrong type for aron double"
        path: list[str] = j["_ARON_PATH"]
        value: np.float64 = np.float64(float(j["_ARON_VALUE"]))
        return cls(path=path, value=value)


@dataclass
class AronString(AronString):
    IDENTIFICATOR = "string"

    value: str = ""
    path: list[str] = field(default_factory=list)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_VALUE": self.value,
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_STRING"
        }
        return j

    def to_primitive(self):
        return self.value

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_STRING", "received wrong type for aron string"
        path: list[str] = j["_ARON_PATH"]
        value: str = j["_ARON_VALUE"]
        return cls(path=path, value=value)


@dataclass
class AronBool(AronBool):
    IDENTIFICATOR = "bool"

    value: bool = False
    path: list[str] = field(default_factory=list)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_VALUE": self.value,
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_BOOL"
        }
        return j

    def to_primitive(self):
        return self.value

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_BOOL", "received wrong type for aron bool"
        path: list[str] = j["_ARON_PATH"]
        value: bool = bool(j["_ARON_VALUE"])
        return cls(path=path, value=value)


# ARON ndarray types
@dataclass
class AronNDArray(AronNDArray):
    IDENTIFICATOR = "ndarray"

    dimensions: list[int] = field(default_factory=list)
    used_type: str = ""
    data: list[int] = field(default_factory=list)
    path: list[str] = field(default_factory=list)
    _array: np.ndarray | None = field(default_factory=lambda: None)  # ONLY SET IF ALREADY CONVERTED

    TYPE_TO_BYTESIZE = {
        "int32": 4,
        "int64": 8,
        "float32": 4,
        "float64": 8,
        "int": 4,  # legacy
        "long": 8,  # legacy
        "float": 4,  # legacy
        "double": 8  # legacy
    }

    DTYPE_TO_TYPE_STRING = {
        "int32": "int32",
        "int64": "int64",
        "float32": "float32",
        "float64": "float64",
        np.int32: "int32",
        np.int64: "int64",
        np.float32: "float32",
        np.float64: "float64"
    }

    TYPE_STRING_TO_DTYPE = {
        "int32": np.int32,
        "int64": np.int64,
        "float32": np.float32,
        "float64": np.float64,
        "int": np.int32,  # legacy
        "long": np.int64,  # legacy
        "float": np.float32,  # legacy
        "double": np.float64  # legacy
    }

    def to_primitive(self):
        return self.toNDArray()

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_DIMESIONS": self.dimensions,
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_NDARRAY",
            "_ARON_USED_TYPE": self.used_type,
            "_ARON_DATA": self.data
        }
        return j

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_NDARRAY", "received wrong type for aron ndarray"
        path: list[str] = j["_ARON_PATH"]
        dims: list[int] = j["_ARON_DIMESIONS"]
        used_type: str = j["_ARON_USED_TYPE"]
        data: list[int] = j["_ARON_DATA"]

        return cls(path=path, dimensions=dims, used_type=used_type, data=data, _array=None)

    def toNumpy(self) -> np.ndarray:
        pass
        # TODO: add conversions to numpy again?
        # return conv.convert(self.data, self.dimensions, conv.TYPE_STRING_TO_DTYPE[self.used_type])

    def toNDArray(self) -> np.ndarray:
        if self._array is not None:
            return self._array

        assert self.data is not None

        # else we need to get rid of the last dimension (bytesize). For convenience we store it as a member
        self._array = self._toNDArray(self.data, self.dimensions, self.TYPE_STRING_TO_DTYPE[self.used_type])
        return self._array

    @classmethod
    def fromNDArray(cls, array: np.ndarray, path=[]):
        by, dims, usedType = cls._fromNDArray(array)

        # for now we ignore the byte-array as it is most likely not used anyway
        return cls(path=path, dimensions=dims, used_type=usedType, data=None, _array=array)

    @classmethod
    def _toNDArray(cls, array: list[int], shape: tuple[int, ...], dtype: np.int32 | np.int64 | np.float32 | np.float64) -> np.ndarray:
        bytesize = shape[-1]
        assert len(array) % bytesize == 0, "len(array) [= " + str(len((array))) + "] must be dividible by " + str(bytesize)
        
        l = []
        index = 0
        while index < len(array):
            bytes = [array[x] for x in range(index, index + bytesize)]
            assert len(bytes) == bytesize, "The size of bytes does not match the given bytesize"
            x = struct.unpack('<f', struct.pack(str(bytesize) + 'B', *bytes))[0]
            index = index + bytesize
            l.append(x)

        data = np.asarray(l, dtype=dtype)
        return np.reshape(data, shape[0:-1])  # we remove the last

    @classmethod
    def _fromNDArray(cls, array: np.ndarray) -> tuple[list[int], list[int], str]:
        by = list(array.tobytes())
        usedType = cls.DTYPE_TO_TYPE_STRING[str(array.dtype)]
        dims = list(array.shape) + [cls.TYPE_TO_BYTESIZE[usedType]]
        return (by, dims, usedType)


# ARON container types
@dataclass
class AronList(AronList):
    IDENTIFICATOR = "list"

    elements: list[AronData] = field(default_factory=list)
    path: list[str] = field(default_factory=list)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_ELEMENTS": [],
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_LIST"
        }

        for k, v in enumerate(self.elements):
            if v is None:
                j["_ARON_ELEMENTS"] += [v]
            else:
                assert isinstance(v, AronData)
                j["_ARON_ELEMENTS"] += [v.toJSON()]

                j["_ARON_ELEMENTS"] += [v.toJSON()]
        return j

    def to_primitive(self):
        return [x.to_primitive() for x in self.elements]

    def at(self, s: int):
        return self.elements[s]

    def atPath(self, path: list[str]) -> AronData:
        if len(path) == 0:
            raise ValueError("Could not access element given empty path")
        if len(path) == 1:
            return self.at(int(path[0]))

        el: AronDict | AronList = self.at(int(path[0]))
        return el.atPath(path[1:])

    def append(self, v: AronData) -> None:
        self.add(len(self.elements), v)

    def add(self, s: int, v: AronData) -> None:
        assert len(self.elements) == s, "Could only add elements to end of list"
        self.elements[s] = v

    def addAtPath(self, path: list[str], v: AronData) -> None:
        if len(path) == 0:
            raise ValueError("Could not add element given empty path")
        if len(path) == 1:
            return self.add(int(path[0]), v)

        el: AronDict | AronList = self.at(int(path[0]))
        return el.addAtPath(path[1:], v)

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_LIST", "received wrong type for aron list"

        elements: list[AronData] = []
        path: list[str] = j["_ARON_PATH"]

        for k, v in enumerate(j["_ARON_ELEMENTS"]):
            if v["_ARON_TYPE"] == "_ARON_INT":
                d = AronInt.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_LONG":
                d = AronLong.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_FLOAT":
                d = AronFloat.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_DOUBLE":
                d = AronDouble.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_STRING":
                d = AronString.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_BOOL":
                d = AronBool.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_NDARRAY":
                d = AronNDArray.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_LIST":
                d = AronList.fromJSON(v)
                elements += [d]
            elif v["_ARON_TYPE"] == "_ARON_DICT":
                d = AronDict.fromJSON(v)
                elements += [d]
            else:
                raise Exception(f"No valid aron type found for {v}")

        return cls(path=path, elements=elements)


@dataclass
class AronDict(AronDict):
    IDENTIFICATOR = "dict"

    elements: dict[str, AronData] = field(default_factory=dict)
    path: list[str] = field(default_factory=list)

    def at(self, s: str) -> AronData:
        return self.elements[s]

    def atPath(self, path: list[str]) -> AronData:
        if len(path) == 0:
            raise ValueError("Could not access element given empty path")
        if len(path) == 1:
            return self.at(path[0])

        el: AronDict | AronList = self.at(path[0])
        return el.atPath(path[1:])

    def add(self, k: str, v: AronData) -> None:
        self.elements[k] = v

    def addAtPath(self, path: list[str], v: AronData) -> None:
        if len(path) == 0:
            raise ValueError("Could not add element given empty path")
        if len(path) == 1:
            return self.add(path[0], v)

        el: AronDict | AronList = self.at(path[0])
        return el.addAtPath(path[1:], v)

    def toJSON(self) -> dict[str, Any]:
        j = {
            "_ARON_ELEMENTS": {},
            "_ARON_PATH": self.path,
            "_ARON_TYPE": "_ARON_DICT"
        }

        for k, v in self.elements.items():
            if k not in j["_ARON_PATH"]:  # otherwise the paths will be copied when loading and storing multiple times
                j["_ARON_PATH"] = self.path + [k]
            else:
                j["_ARON_PATH"] = self.path

            if v is None:
                j["_ARON_ELEMENTS"][k] = v
            else:
                assert isinstance(v, AronData)
                j["_ARON_ELEMENTS"][k] = v.toJSON()

        return j

    def to_primitive(self):
        return {k: v.to_primitive() for k, v in self.elements.items()}

    @classmethod
    def fromJSON(cls, j: dict[str, Any]):
        assert j["_ARON_TYPE"] == "_ARON_DICT", "received wrong type for aron dict"

        elements: dict[str, Any] = {}
        path: list[str] = j["_ARON_PATH"]

        for k, v in j["_ARON_ELEMENTS"].items():
            if v is None:
                # null (e.g. not set optional)
                # ignore and hope that its never set
                pass
            # elif "image" in k:
            # fix for now because element in here should be null but is not due to multiple LTM recordings
            # pass
            elif v["_ARON_TYPE"] == "_ARON_INT":
                d = AronInt.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_LONG":
                d = AronLong.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_FLOAT":
                d = AronFloat.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_DOUBLE":
                d = AronDouble.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_STRING":
                d = AronString.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_BOOL":
                d = AronBool.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_NDARRAY":
                d = AronNDArray.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_LIST":
                d = AronList.fromJSON(v)
                elements[k] = d
            elif v["_ARON_TYPE"] == "_ARON_DICT":
                d = AronDict.fromJSON(v)
                elements[k] = d
            else:
                raise Exception("No valid aron type found. Received: " + v["_ARON_TYPE"])

        return cls(path=path, elements=elements)
