import numpy as np

from armarx_memory.aron.data.variant import AronData, AronBool, AronInt, AronDict, AronDouble, AronFloat, AronList, AronLong, \
    AronNDArray, AronString
from armarx_memory.aron.data.dto.common.instantiated_type import InstantiatedLeafType, InstantiatedContainerType


# Not really a visitor. Will be moved probably.make factory
class BackConverter:

    def __init__(self, n: np.ndarray, type_structure: InstantiatedContainerType) -> None:
        assert n.shape[-1] == 1

        self.array: np.ndarray = n
        self.type_structure: InstantiatedContainerType = type_structure
        self.next_index: int = 0

    def createBool(self, path: list[str]) -> AronBool:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronBool.IDENTIFICATOR
        assert s.cols == 1

        d = AronBool(path=path)
        d.value = bool(self.array[self.next_index, 0])
        self.next_index += 1
        return d

    def createInt(self, path: list[str]) -> AronInt:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronInt.IDENTIFICATOR
        assert s.cols == 1

        d = AronInt(path=path)
        d.value = np.int32(self.array[self.next_index, 0])
        self.next_index += 1
        return d

    def createLong(self, path: list[str]) -> AronLong:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronLong.IDENTIFICATOR
        assert s.cols == 1
        
        d = AronLong(path=path)
        d.value = np.int64(self.array[self.next_index, 0])
        self.next_index += 1
        return d

    def createFloat(self, path: list[str]) -> AronFloat:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronFloat.IDENTIFICATOR
        assert s.cols == 1
        
        d = AronFloat(path=path)
        d.value = np.float32(self.array[self.next_index, 0])
        self.next_index += 1
        return d

    def createDouble(self, path: list[str]) -> AronDouble:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronDouble.IDENTIFICATOR
        assert s.cols == 1
        
        d = AronDouble(path=path)
        d.value = np.float64(self.array[self.next_index, 0])
        self.next_index += 1
        return d

    def createString(self, path: list[str]) -> AronString:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronString.IDENTIFICATOR
        
        d = AronString(path=path)  # for now we ignore strings
        return d

    def createNDArray(self, path: list[str]) -> AronNDArray:
        s = self.type_structure.at(path)
        assert isinstance(s, InstantiatedLeafType)
        assert s.type == AronNDArray.IDENTIFICATOR
        
        start = self.next_index 
        end = self.next_index + s.cols

        x = self.array[start:end]
        assert x.shape[-1] == 1

        d = AronNDArray.fromNDArray(np.reshape(x, x.shape[:-1]), path=path)
        #d.dimensions = list((s.cols,) + self.array.shape[1:-1] + (4,)) # remove first and last dimension
        #d.used_type = s.additional_info
        #d.data = self.array[start:end, 0]
        self.next_index = end
        return d


    def createDict(self, path: list[str]) -> AronDict:
        s = self.type_structure
        if len(path) != 0:
            s = self.type_structure.at(path)

        assert isinstance(s, InstantiatedContainerType)
        assert isinstance(s.elements, dict)
        assert s.type == AronDict.IDENTIFICATOR

        d = AronDict(path=path)

        for k, i in s.elements.items():
            e = self.create(path + [k])
            d.add(k, e)

        return d
    
    def createList(self, path: list[str]) -> AronList:
        s = self.type_structure
        if len(path) != 0:
            s = self.type_structure.at(path)

        assert isinstance(s, InstantiatedContainerType)
        assert isinstance(s.elements, list)
        assert s.type == AronList.IDENTIFICATOR

        d = AronList(path=path)

        for i in range(len(s.elements)):
            e = self.create(path + [str(i)])
            d.append(e)

        return d
    
    
    def create(self, path=[]) -> AronData:
        s = self.type_structure
        if len(path) != 0:
            s = self.type_structure.at(path)

        if isinstance(s, InstantiatedLeafType):
            t = s.type
            if t == AronInt.IDENTIFICATOR:
                return self.createInt(path)
            elif t == AronLong.IDENTIFICATOR:
                return self.createLong(path)
            elif t == AronFloat.IDENTIFICATOR:
                return self.createFloat(path)
            elif t == AronDouble.IDENTIFICATOR:
                return self.createDouble(path)
            elif t == AronString.IDENTIFICATOR:
                return self.createString(path)
            elif t == AronBool.IDENTIFICATOR:
                return self.createBool(path)
            elif t == AronNDArray.IDENTIFICATOR:
                return self.createNDArray(path)
            else:
                raise ValueError("Got unknown leaf type " + t)
        elif isinstance(s, InstantiatedContainerType):
            t = s.type
            if t == AronList.IDENTIFICATOR:
                return self.createList(path)
            elif t == AronDict.IDENTIFICATOR:
                return self.createDict(path)
            else: ValueError("Got unknown container type " + t)
        else:
            raise ValueError("Unknown instantiated type")