import numpy as np

from armarx_memory.aron.data.variant import (AronBool, AronInt, AronDouble, AronFloat, AronLong, AronNDArray, AronString, AronDict,
                                  AronList)
from armarx_memory.aron.data.dto.common.instantiated_type import InstantiatedLeafType, InstantiatedContainerType
from armarx_memory.aron.data.visitors.visitor import RecursiveVisitorInterface


class Converter(RecursiveVisitorInterface):

    def __init__(self, shape: tuple[int, ...]) -> None:
        assert shape[0] == -1, "Requires a dynamic shape using -1"
        assert shape[-1] == 1, "Requires 1 as last dimension!"

        self.shape = shape
        self.array: np.ndarray = np.empty((0,) + shape[1:], dtype=np.float32)
        self.type_structure: InstantiatedContainerType = InstantiatedContainerType(type=AronDict.IDENTIFICATOR, cols=-1, elements={})

        self._cols_stack = []

    def visitBool(self, d: AronBool) -> None:
        assert d.IDENTIFICATOR == AronBool.IDENTIFICATOR
        assert len(self._cols_stack) > 0

        arr = np.array([d.value], dtype=np.float32)
        r = np.reshape(arr, self.shape)
        self.array = np.append(self.array, r, axis=0)

        leaf_type_structure = InstantiatedLeafType(type=AronBool.IDENTIFICATOR, cols=r.shape[0])
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = leaf_type_structure

        self._cols_stack[-1] += r.shape[0]

    def visitInt(self, d: AronInt) -> None:
        assert d.IDENTIFICATOR == AronInt.IDENTIFICATOR
        assert len(self._cols_stack) > 0

        arr = np.array([d.value], dtype=np.float32)
        r = np.reshape(arr, self.shape)
        self.array = np.append(self.array, r, axis=0)

        leaf_type_structure = InstantiatedLeafType(type=AronInt.IDENTIFICATOR, cols=r.shape[0])
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = leaf_type_structure

        self._cols_stack[-1] += r.shape[0]

    def visitLong(self, d: AronLong) -> None:
        assert d.IDENTIFICATOR == AronLong.IDENTIFICATOR
        assert len(self._cols_stack) > 0
        
        arr = np.array([d.value], dtype=np.float32)
        r = np.reshape(arr, self.shape)
        self.array = np.append(self.array, r, axis=0)
        
        leaf_type_structure = InstantiatedLeafType(type=AronLong.IDENTIFICATOR, cols=r.shape[0])
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = leaf_type_structure

        self._cols_stack[-1] += r.shape[0]

    def visitFloat(self, d: AronFloat) -> None:
        assert d.IDENTIFICATOR == AronFloat.IDENTIFICATOR
        assert len(self._cols_stack) > 0
        
        arr = np.array([d.value], dtype=np.float32)
        r = np.reshape(arr, self.shape)
        self.array = np.append(self.array, r, axis=0)
        
        leaf_type_structure = InstantiatedLeafType(type=AronFloat.IDENTIFICATOR, cols=r.shape[0])
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = leaf_type_structure
        
        self._cols_stack[-1] += r.shape[0]

    def visitDouble(self, d: AronDouble) -> None:
        assert d.IDENTIFICATOR == AronDouble.IDENTIFICATOR
        assert len(self._cols_stack) > 0
        
        arr = np.array([d.value], dtype=np.float32)
        r = np.reshape(arr, self.shape)
        self.array = np.append(self.array, r, axis=0)
        
        leaf_type_structure = InstantiatedLeafType(type=AronDouble.IDENTIFICATOR, cols=r.shape[0])
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = leaf_type_structure
        
        self._cols_stack[-1] += r.shape[0]

    def visitString(self, d: AronString) -> None:
        assert d.IDENTIFICATOR == AronString.IDENTIFICATOR
        assert len(self._cols_stack) > 0
        
        pass # we ignore strings

    def visitNDArray(self, d: AronNDArray) -> None:
        assert d.IDENTIFICATOR == AronNDArray.IDENTIFICATOR
        assert len(self._cols_stack) > 0
        
        arr = d.toNDArray()
        r = np.reshape(arr, self.shape)
        self.array = np.append(self.array, r, axis=0)
        
        leaf_type_structure = InstantiatedLeafType(type=AronNDArray.IDENTIFICATOR, cols=r.shape[0], additional_info=d.used_type)
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = leaf_type_structure

        self._cols_stack[-1] += r.shape[0]

    def visitDictStart(self, d: AronDict) -> None:
        assert d.IDENTIFICATOR == AronDict.IDENTIFICATOR

        container_type_structure = InstantiatedContainerType(type=AronDict.IDENTIFICATOR, cols=-1, elements={})

        if len(d.path) > 0:
            self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = container_type_structure
        else:
            self.type_structure = container_type_structure

        self._cols_stack += [0]

    def visitDictEnd(self, d: AronDict) -> None:
        assert d.IDENTIFICATOR == AronDict.IDENTIFICATOR
        assert len(self._cols_stack) > 0

        self.type_structure.at(d.path).cols = self._cols_stack[-1]
        self._cols_stack = self._cols_stack[:-1]


    def visitListStart(self, d: AronList) -> None:
        assert d.IDENTIFICATOR == AronList.IDENTIFICATOR
        
        container_type_structure = InstantiatedContainerType(type=AronList.IDENTIFICATOR, cols=-1, elements=[])
        self.type_structure.at(d.path[:-1]).elements[d.path[-1]] = container_type_structure

        self._cols_stack += [0]

    def visitListEnd(self, d: AronList) -> None:
        assert d.IDENTIFICATOR == AronList.IDENTIFICATOR
        assert len(self._cols_stack) > 0

        self.type_structure.at(d.path).cols = self._cols_stack[-1]

        self._cols_stack = self._cols_stack[:-1]