import os
from typing import TypeVar, Generic, get_args
import numpy as np

from armarx_memory.core.MemoryID import MemoryID
from armarx_memory.ltm.base.entity_snapshot import EntitySnapshot
from armarx_memory.ltm.util.filesystem_handler import escapeName, unescapeName
from armarx_memory.ltm.util.timestamp_conversions import date_from_ts, seconds_from_ts

EntitySnapshotType = TypeVar('EntitySnapshotType')


class Entity(Generic[EntitySnapshotType]):

    TIMESTAMP_NORMALIZATION = 1.0 / 1000000.0  # bring ts to seconds

    def __init__(self, id: MemoryID, basepath: str):
        self.id: MemoryID = id.getEntityID()
        self.basepath: str = basepath
        self.rootdir = self.basepath + "/" + escapeName(self.id.entity_name)
        self.snapshots: dict[int, EntitySnapshotType] = {}

    def loadReferences(self):
        subdirsDays = [f.path for f in os.scandir(self.rootdir) if f.is_dir()]

        for subdirDay in subdirsDays:
            subdirsSeconds = [f.path for f in os.scandir(subdirDay) if f.is_dir()]
            for subdirSecond in subdirsSeconds:
                subdirsMicroSeconds = [f.name for f in os.scandir(subdirSecond) if f.is_dir()]
                for subdirMicroSecondName in subdirsMicroSeconds:
                    ts = int(subdirMicroSecondName)

                    # only load what's necessary
                    if self.id.hasTimestamp() and self.id.timestamp_usec != ts:
                        continue

                    self.addEntitySnapshot(self.id.changeTimestamp(ts))
                    self.snapshots[ts].loadReferences()

    def addEntitySnapshot(self, id: MemoryID = None, entitySnapshot: EntitySnapshot = None):
        if id != None and entitySnapshot != None:
            assert id == entitySnapshot.id, "You are trying to add a core segment to the memory but the id given and the id of the core segment object do not match"
        if entitySnapshot != None:
            self.snapshots[entitySnapshot.id.timestamp_usec] = entitySnapshot
        elif id != None:
            date_dir: str = date_from_ts(id.timestamp_usec, self.TIMESTAMP_NORMALIZATION)
            date_seconds: int = seconds_from_ts(id.timestamp_usec, self.TIMESTAMP_NORMALIZATION)

            rootDir = self.rootdir + "/" + date_dir + "/" + str(date_seconds)

            snapshot = EntitySnapshot(id, rootDir)
            self.snapshots[id.timestamp_usec] = snapshot
        else:
            assert False, "You are trying to add a snapshot to this entity, but you have neither provided an id nor a snapshot object"
        

    def store(self):
        print(self.rootdir)
        os.makedirs(self.rootdir, exist_ok=True)

        for k, s in self.snapshots.items():
            pp = (self.rootdir + "/" + date_from_ts(k, self.TIMESTAMP_NORMALIZATION) + "/" +
                  str(seconds_from_ts(k, self.TIMESTAMP_NORMALIZATION)) + "/")
            os.makedirs(pp, exist_ok=True)
            s.store()

    def for_each_snapshot(self, fn):
        for _, snapshot in self.snapshots.items():
            fn(snapshot)

    def for_each_instance(self, fn):
        for _, snapshot in self.snapshots.items():
            snapshot.for_each_instance(fn)

    def for_each_data(self, fn):
        for _, snapshot in self.snapshots.items():
            snapshot.for_each_data(fn)

    # TODO: to_ndarray and from_ndarray for entities

    """

    def to_ndarray(self, shape: tuple[int, ...] = (0, 1)) -> dict[int, tuple[np.ndarray, list[InstantiatedContainerType]]]:
        d: dict[int, tuple[np.ndarray, list[InstantiatedContainerType]]] = {}

        for k, snapshot in self.snapshots.items():
            i = snapshot.to_ndarray(shape)
            d[k] = i

        return d
    
    def from_ndarray(cls, entity_id: MemoryID, basepath: str, data: dict[int, tuple[np.ndarray, list[InstantiatedContainerType]]]) -> None:
        entity: cls = cls(entity_id, basepath)
        T = get_args(cls.__orig_bases__[0])[0]

        for timestamp, (array, type_structures) in data.items():
            pp = (entity.rootdir + "/" + date_from_ts(timestamp, cls.TIMESTAMP_NORMALIZATION) + "/" + 
                  str(seconds_from_ts(timestamp, cls.TIMESTAMP_NORMALIZATION)) + "/")

            snapshot = T.from_ndarray(entity_id.changeTimestamp(timestamp), pp, array, type_structures)
            entity.snapshots[timestamp] = snapshot
        return entity
        
    """

    def __len__(self):
        return len(self.snapshots)

