import os
import json
from typing import TypeVar, Generic
from PIL import Image

import numpy as np
from armarx_memory.aron.data.dto.common.instantiated_type import InstantiatedContainerType
from armarx_memory.aron.data.dto.common.metadata import Metadata
from armarx_memory.aron.data.variant import AronNDArray
from armarx_memory.aron.data.visitors.ndarray.shape_calculator import ShapeCalculator as NDArrayShapeCalculator
from armarx_memory.aron.data.visitors.ndarray.converter import Converter as NDArrayConverter
from armarx_memory.aron.data.visitors.ndarray.back_converter import BackConverter as NDArrayBackConverter

from armarx_memory.core.MemoryID import MemoryID

from armarx_memory.aron.data.variant import AronDict

DataType = TypeVar('DataType')
MetadataType = TypeVar('MetadataType')


class EntityInstance(Generic[DataType, MetadataType]):

    DATA_FILENAME = "data.aron.json"
    METADATA_FILENAME = "metadata.aron.json"

    def __init__(self, id: MemoryID, basepath: str):
        self.id: MemoryID = id
        self.basepath: str = basepath
        self.data: DataType | None = None
        self.metadata: MetadataType | None = None

        self.rootdir = self.basepath + "/" + str(self.id.instance_index)

    def loadReferences(self):
        metadataPath = self.rootdir + "/" + self.METADATA_FILENAME
        if os.path.exists(metadataPath):
            with open(metadataPath) as f:
                j = json.load(f)
                m = AronDict.fromJSON(j)
                self.metadata = Metadata.fromAron(m)

    def load(self, reload=False):
        if self.data is not None and not reload:
            # Data has already been loaded.
            return

        dataPath = self.rootdir + "/" + self.DATA_FILENAME

        self.loadReferences()

        if os.path.exists(dataPath):
            with open(dataPath) as f:
                json_data = json.load(f)
                self.data = AronDict.fromJSON(json_data)

                # resolve special exports
                pngs = [_f.name for _f in os.scandir(self.rootdir) if _f.is_file() and _f.name.endswith(".png") and
                        not ("detected" in _f.name)]
                for png in pngs:
                    basename = os.path.basename(png)
                    stem = basename.replace(".rgb.png", "")
                    image = Image.open(self.rootdir + "/" + png)
                    data = np.asarray(image, dtype=np.float32)
                    data /= 255.0

                    by, dims, usedtype = AronNDArray._fromNDArray(data)
                    self.data.elements[stem] = AronNDArray(path=[stem], dimensions=dims, used_type=usedtype, data=None,
                                                           _array=data)

        

    def store(self):
        dataPath = self.rootdir + "/" + self.DATA_FILENAME
        metadataPath = self.rootdir + "/" + self.METADATA_FILENAME

        os.makedirs(self.rootdir, exist_ok=True)

        if self.data is not None:
            for k, v in self.data.elements.items():
                if k == "image":
                    assert isinstance(v, AronNDArray)
                    image = Image.fromarray(np.uint8(v._array * 255))
                    image.save(self.rootdir + "/" + k + ".rgb.png")

            j = self.data.toJSON()
            with open(dataPath, "w") as f:
                json.dump(j, f, indent=4)

        if self.metadata is not None:
            j = self.metadata.toJSON()
            with open(metadataPath, "w") as f:
                json.dump(j, f, indent=4)

    def for_each_data(self, fn):
        fn(self.id, self.data)

    def for_each_metadata(self, fn):
        fn(self.id, self.metadata)

    # TODO: get_shape, to_ndarray and from_ndarray for entity instances

    def get_shape(self) -> tuple[int, ...]:
        self.load(False)

        c: NDArrayShapeCalculator = NDArrayShapeCalculator()
        c.visit(self.data)
        return c.shape
    
    def to_ndarray(self, shape: tuple[int, ...]) -> tuple[np.ndarray, InstantiatedContainerType]:
        self.load(False)  # assure that data is loaded

        c: NDArrayConverter = NDArrayConverter(shape)
        c.visit(self.data)

        #self.data = None  # free space

        return c.array, c.type_structure

    @staticmethod
    def from_ndarray(instance_id: MemoryID, basepath: str, array: np.ndarray, type_structure: InstantiatedContainerType):
        # in from aron there is no way to only store a reference as we do not know where the data is stored...
        c: NDArrayBackConverter = NDArrayBackConverter(array, type_structure)
        aron: AronDict = c.create()

        data = aron
        id = instance_id
        new_instance = EntityInstance(id, basepath)
        new_instance.data = data

        return new_instance

    # TODO: make sure that from and to ndarray can be used as many times as the user wants without adding any dimensions
