import os
from typing import TypeVar, Generic, get_args
import numpy as np
from armarx_memory.aron.data.dto.common.instantiated_type import InstantiatedContainerType

from armarx_memory.core.MemoryID import MemoryID
from armarx_memory.ltm.base.entity_instance import EntityInstance

EntityInstanceType = TypeVar('EntityInstanceType')


class EntitySnapshot(Generic[EntityInstanceType]):

    def __init__(self, id: MemoryID, basepath: str):
        self.id: MemoryID = id.getSnapshotID()
        self.basepath: str = basepath
        self.rootdir = self.basepath + "/" + str(self.id.timestamp_usec)
        self.instances: list[EntityInstanceType] = []

    def loadReferences(self):
        subdirs = [f.name for f in os.scandir(self.rootdir) if f.is_dir()]

        for subdirName in sorted(subdirs):
            index = int(subdirName)

            # only load what's necessary
            if self.id.hasIndex() and self.id.instance_index != index:
                continue

            self.addEntityInstance(self.id.changeIndex(index))
            self.instances[index].loadReferences()

    def addEntityInstance(self, id: MemoryID = None, entityInstance: EntityInstance = None):
        if id != None and entityInstance != None:
            assert id == entityInstance.id, "You are trying to add an instance to the snapshot but the id given and the id of the instance object do not match"
        if entityInstance != None:
            self.instances += [entityInstance]
        elif id != None:
            instance = EntityInstance(id, self.rootdir)
            self.instances += [instance]
        else:
            assert False, "You are trying to add an Entity Instance to this Snapshot, but you have neither provided an id nor an Instance object"

    def store(self):
        os.makedirs(self.rootdir, exist_ok=True)

        for i in self.instances:
            i.store()

    def for_each_instance(self, fn):
        for instance in self.instances:
            fn(instance)

    def for_each_data(self, fn):
        for instance in self.instances:
            instance.for_each_data(fn)

    def get_shape(self) -> tuple[int, ...]:
        shape = None
        for instance in self.instances:
            if shape is None:
                shape = instance.to_shape()
            else:
                s = instance.to_shape()
                assert shape[1:] == s[1:], "The shapes od two instances do not match!!"

                shape = (shape[0] + s[0],) + shape[1:]

        assert shape[-1] == 1, "Last element of shape must be 1"
        return shape  # returns actual shape (incl full amount of rows as first element)

    # TODO: to_ndarray and from_ndarray are not correct yet, to_ndarray correctly adds the second instance to the array
    # but it vanishes somewhere

    def to_ndarray(self, shape: tuple[int, ...]) -> tuple[np.ndarray, list[InstantiatedContainerType]]:
        assert shape[0] == -1, "First element of shape must be -1"
        assert shape[-1] == 1, "Last element of shape must be 1"

        array = np.empty((0,) + shape[1:], dtype=np.float32)
        type_structure: list[InstantiatedContainerType] = []
        for instance in self.instances:
            instance_array, instance_type_structure = instance.to_ndarray(shape)
            array = np.append(array, instance_array, axis=0)
            type_structure += [instance_type_structure]
        return array, type_structure


    @staticmethod
    def from_ndarray(snapshot_id: MemoryID, basepath: str, array: np.ndarray, type_structure_list: list[InstantiatedContainerType]):
        current_index = 0

        snapshot = EntitySnapshot(snapshot_id, basepath)

        for i, type_structure in enumerate(type_structure_list):
            if type_structure.type != 'ndarray':
                # in this case the type_structure is a container and holds a ndarray somewhere deeper
                type_structures = type_structure.elements.values()
                if len(type_structures) == 1:
                    overall_size = list(type_structures)[0].cols
                else:  # this case might be wrong
                    overall_size = -1
            else:
                overall_size = type_structure.cols
            end_index = current_index + overall_size

            ins = EntityInstance.from_ndarray(snapshot_id.changeIndex(i), snapshot.rootdir,
                                              array[current_index:end_index], type_structure)
            snapshot.instances += [ins]
            if overall_size == -1:
                # in this case the end index was -1 and not the correct amount of elements
                current_index += len(array[current_index:end_index])
                pass
            else:
                current_index = end_index
        return snapshot


    def len(self):
        return len(self.instances)
