import dataclasses as dc
import typing as ty

from armarx_memory.aron.aron_dataclass import AronDataclass
from armarx_memory.aron.common.names import Names
from armarx_memory.aron.common.framed import FramedPose
from armarx_memory.core import MemoryID
from armarx_memory.client.Reader import Reader
from armarx_memory.client.MemoryNameSystem import MemoryNameSystem

from armarx_core.tools import fuzzy_multi_match


CORE_SEGMENT_ID = MemoryID("Navigation", "Location")

@dc.dataclass
class Location(AronDataclass):

    framed_pose: FramedPose
    names: ty.Optional[Names]

    @classmethod
    def _get_conversion_options(cls) -> ty.Optional["ConversionOptions"]:
        from armarx_memory.aron.conversion import ConversionOptions
        return ConversionOptions(names_snake_case_to_camel_case=True)


@dc.dataclass
class ResolvedLocationName(AronDataclass):

    location_id: MemoryID
    location_name: str
    match_ratio: float


class LocationReader:

    def __init__(self, location_segment_reader: Reader):
        self.reader = location_segment_reader

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem, wait=True) -> "LocationReader":
        return cls(
            mns.wait_for_reader(CORE_SEGMENT_ID)
            if wait
            else mns.get_reader(CORE_SEGMENT_ID)
        )

    def resolve_location_name(
            self,
            location_name: str,
            min_match_ratio = 0.5,
            log=None,
    ) -> ty.Optional[ResolvedLocationName]:
        # Get locations from memory
        locations = self.reader.query_latest(CORE_SEGMENT_ID)

        location_ids_to_words = dict()

        def instance_fn(id: MemoryID, data: ty.Dict):
            location = Location.from_dict(data)

            location_ids_to_words[id] = [id.entity_name]
            if location.names is not None:
                location_ids_to_words[id] += location.names.recognized
            # TODO: Split entityName by "/" and add components?

        self.reader.for_each_instance_data(instance_fn, locations)

        if len(location_ids_to_words) == 0:
            return None

        match = fuzzy_multi_match.fuzzy_multi_match(location_name, location_ids_to_words, log=log)
        if match.ratio < min_match_ratio:
            return None

        return ResolvedLocationName(
            location_id=match.option,
            location_name=match.word,
            match_ratio=match.ratio,
        )
