import dataclasses as dc
from typing import List, Optional

from armarx_memory.aron.aron_dataclass import AronDataclass
from armarx_memory.client.detail.SpecialClientBase import SpecialWriterBase, SpecialReaderBase
from armarx_memory.core import MemoryID

CORE_SEGMENT_ID = MemoryID("Speech", "SpeechToText")


@dc.dataclass
class SpeechToText(AronDataclass):
    text: str

    speaker: Optional[MemoryID] = None
    """
    This can be set to a MemoryID pointing to Human/Profile, identifying the speaker of this recognized text.
    Only set if the running ASR supports speaker recognition.  
    """


class SpeechToTextWriter(SpecialWriterBase):
    __default_entity_name__ = 'text'

    @classmethod
    def _get_aron_class(cls):
        return SpeechToText

    @classmethod
    def _get_default_core_segment_id(cls) -> MemoryID:
        return CORE_SEGMENT_ID


class SpeechToTextReader(SpecialReaderBase):
    @classmethod
    def _get_aron_class(cls):
        return SpeechToText

    @classmethod
    def _get_default_core_segment_id(cls) -> MemoryID:
        return CORE_SEGMENT_ID

    def fetch_latest_instance(self, updated_ids: Optional[List[MemoryID]] = None):
        """
        Query the latest snapshot of the given updated IDs and
        return its first instance.
        """
        if updated_ids is None:
            memory = self.reader.query_latest(self.core_segment_id)

            latest_snapshot = None

            core_seg = memory.coreSegments[self.core_segment_id.core_segment_name]
            for prov_seg in core_seg.providerSegments.values():
                for entity in prov_seg.entities.values():
                    for snapshot in entity.history.values():
                        if latest_snapshot is None:
                            latest_snapshot = snapshot
                        elif (
                                latest_snapshot.id.timestamp.timeSinceEpoch.microSeconds
                                < snapshot.id.timestamp.timeSinceEpoch.microSeconds
                        ):
                            latest_snapshot = snapshot
        else:
            for up_id in updated_ids:
                assert self.core_segment_id.contains(up_id)

            latest_snapshot_id = max(updated_ids, key=lambda i: i.timestamp_usec)
            latest_snapshot = self.reader.query_snapshot(latest_snapshot_id)

        if latest_snapshot is not None:
            latest_instance = latest_snapshot.instances[0]
        else:
            latest_instance = None
        return latest_instance
